package org.fictus

class AnyRefConverter[A <: AnyRef](expected: A)(implicit context: MockContext) {
    def same = context.same(expected)
}
class AnyConverter[A](expected: A)(implicit context: MockContext) {
    def equal = context.equal(expected)
}

trait Matchers {
    implicit def forAny[A](expected: A)(implicit context: MockContext) = new AnyConverter(expected)
    implicit def forAnyRef[A <: AnyRef](expected: A)(implicit context: MockContext) = new AnyRefConverter(expected)

    def equal[A](expected: A)(implicit context: MockContext): A = {
        context.currentMatchers.append(Equal(expected))
        expected
    }

    def same[A <: AnyRef](expected: A)(implicit context: MockContext): A = {
        context.currentMatchers.append(Same(expected))
        expected
    }

    def matches[A](expected: A => Boolean)(implicit context: MockContext, manifest: Manifest[A]): A = {
        context.currentMatchers.append(Matches(expected))
        Types.defaultValue(manifest.erasure).asInstanceOf[A]
    }

    // TODO the syntax for this is problematic.
    // using * with no type parameter leads to the type being Object
    // using *[type] works, but is ugly
    // this is relevant http://stackoverflow.com/questions/9118264/multiple-parameter-closure-argument-type-not-inferred
    def *[A](implicit context: MockContext): A = any

    def any[A](implicit context: MockContext): A = {
        val default = Types.defaultValue(manifest.erasure).asInstanceOf[A]
        context.currentMatchers.append(Anything[A](default))
        default
    }

    def createCapture[A]() = Capture[A]()
    def capture[A](captured: Capture[A])(implicit context: MockContext): A = {
        val default = Types.defaultValue(manifest.erasure).asInstanceOf[A]
        context.currentMatchers.append(captured)
        default
    }
}

trait Matcher {
    def matches(actual: Any): Boolean
    def actualMsg(actual: Any) = Option(actual).getOrElse("null").toString
    def expectedMsg: String
}

case class Equal[A](expected: A) extends Matcher {
    override def matches(actual: Any) = expected == actual
    override def expectedMsg = Option(expected).getOrElse("null").toString
}

case class Same[A <: AnyRef](expected: A) extends Matcher {
    override def matches(actual: Any) = actual match {
        case actual: AnyRef => expected eq actual
        case _              => false
    }
    override def expectedMsg = Option(expected).getOrElse("null").toString
}

case class Matches[A](expected: A => Boolean) extends Matcher {
    override def matches(actual: Any) = expected(actual.asInstanceOf[A])
    override def actualMsg(actual: Any) = "Matches: " + matches(actual)
    override def expectedMsg = "Matches: true"
}

case class Anything[A](expected: A) extends Matcher {
    override def matches(actual: Any) = true
    override def expectedMsg = "*"
    override def actualMsg(actual: Any) = "*"
}

case class Capture[A](var value: Option[A] = None) extends Matcher {
    override def matches(actual: Any) = {
        value = Some(actual.asInstanceOf[A])
        true
    }
    override def expectedMsg = "*"
    override def actualMsg(actual: Any) = "*"
}
